# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from typing import TYPE_CHECKING

# WARNING: do not import unnecessary things here to keep cli startup time under
# control


if TYPE_CHECKING:
    from typing import Any, Dict, Optional


def read_config(config_file: Optional[Any] = None) -> Dict:
    """Read configuration from config_file if provided, from the SWH_CONFIG_FILENAME if
    set or fallback to the DEFAULT_CONFIG.

    """
    from os import environ

    from swh.core import config
    from swh.scheduler import DEFAULT_CONFIG, DEFAULT_CONFIG_RAW

    if config_file:
        conf = config.read(config_file, DEFAULT_CONFIG)
    elif "SWH_CONFIG_FILENAME" in environ:
        conf = config.load_from_envvar(DEFAULT_CONFIG_RAW)
    else:
        conf = config.read(None, DEFAULT_CONFIG)
    return conf



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This is the scheduler listener. It is in charge of listening to rabbitmq events (the
task result) and flushes the "oneshot" tasks' status in the scheduler backend. It's the
final step after a task is done.

The scheduler runner :mod:`swh.scheduler.celery_backend.runner` is the module in charge
of pushing tasks in the queue.

"""

import json
import logging
import sys

import pika

from swh.core.statsd import statsd
from swh.scheduler import get_scheduler
from swh.scheduler.utils import utcnow

logger = logging.getLogger(__name__)


def get_listener(broker_url, queue_name, scheduler_backend):
    connection = pika.BlockingConnection(pika.URLParameters(broker_url))
    channel = connection.channel()

    channel.queue_declare(queue=queue_name, durable=True)

    exchange = "celeryev"
    routing_key = "#"
    channel.queue_bind(queue=queue_name, exchange=exchange, routing_key=routing_key)

    channel.basic_qos(prefetch_count=1000)

    channel.basic_consume(
        queue=queue_name,
        on_message_callback=get_on_message(scheduler_backend),
    )

    return channel


def get_on_message(scheduler_backend):
    def on_message(channel, method_frame, properties, body):
        try:
            events = json.loads(body)
        except Exception:
            logger.warning("Could not parse body %r", body)
            events = []

        if not isinstance(events, list):
            events = [events]

        for event in events:
            logger.debug("Received event %r", event)
            process_event(event, scheduler_backend)

        channel.basic_ack(delivery_tag=method_frame.delivery_tag)

    return on_message


def process_event(event, scheduler_backend):
    uuid = event.get("uuid")
    if not uuid:
        return

    event_type = event["type"]
    statsd.increment(
        "swh_scheduler_listener_handled_event_total", tags={"event_type": event_type}
    )

    if event_type == "task-started":
        scheduler_backend.start_task_run(
            uuid,
            timestamp=utcnow(),
            metadata={"worker": event.get("hostname")},
        )
    elif event_type == "task-result":
        result = event["result"]
        status = None
        # XXX we probably should expect result to always be a dict at this point...
        if isinstance(result, dict):
            status = result.get("status")
            if status == "success":
                status = "eventful" if result.get("eventful") else "uneventful"

        if status is None:
            # XXX when is this supposed to be used?
            status = "eventful" if result else "uneventful"

        metadata = {}
        if status == "failed" and isinstance(result, dict):
            if "error" in result:
                metadata["error"] = result["error"]

        scheduler_backend.end_task_run(
            uuid, timestamp=utcnow(), status=status, metadata=metadata
        )

    elif event_type == "task-failed":
        scheduler_backend.end_task_run(uuid, timestamp=utcnow(), status="failed")


if __name__ == "__main__":
    url = sys.argv[1]
    logging.basicConfig(level=logging.DEBUG)
    scheduler_backend = get_scheduler(
        "postgresql", args={"db": "service=swh-scheduler"}
    )
    channel = get_listener(url, "celeryev.test", scheduler_backend)
    logger.info("Start consuming")
    channel.start_consuming()



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Tuple
from unittest.mock import MagicMock

import pytest

from swh.scheduler.cli.origin import format_origins
from swh.scheduler.tests.common import TASK_TYPES
from swh.scheduler.tests.test_cli import invoke as basic_invoke


def invoke(
    scheduler, args: Tuple[str, ...] = (), catch_exceptions: bool = False, **kwargs
):
    return basic_invoke(
        scheduler, args=["origin", *args], catch_exceptions=catch_exceptions, **kwargs
    )


def test_cli_origin(swh_scheduler):
    """Check that swh scheduler origin returns its help text"""

    result = invoke(swh_scheduler)

    assert "Commands:" in result.stdout


def test_format_origins_basic(listed_origins):
    listed_origins = listed_origins[:100]

    ctx = MagicMock()

    basic_output = list(format_origins(ctx, listed_origins))
    assert ctx.method_calls == []
    # 1 header line + all origins
    assert len(basic_output) == len(listed_origins) + 1

    no_header_output = list(format_origins(ctx, listed_origins, with_header=False))
    assert ctx.method_calls == []
    assert basic_output[1:] == no_header_output


def test_format_origins_fields_unknown(listed_origins):
    listed_origins = listed_origins[:10]

    ctx = MagicMock()
    it = format_origins(ctx, listed_origins, fields=["unknown_field"])
    assert ctx.method_calls == []

    next(it)
    ctx.fail.assert_called()


def test_format_origins_fields(listed_origins):
    listed_origins = listed_origins[:10]
    fields = ["lister_id", "url", "visit_type"]

    ctx = MagicMock()
    output = list(format_origins(ctx, listed_origins, fields=fields))
    assert ctx.method_calls == []
    assert output[0] == ",".join(fields)
    for i, origin in enumerate(listed_origins):
        assert output[i + 1] == f"{origin.lister_id},{origin.url},{origin.visit_type}"


def test_grab_next(swh_scheduler, listed_origins_by_type):
    NUM_RESULTS = 10
    # Strict inequality to check that grab_next_visits doesn't return more
    # results than requested

    # XXX: should test all of 'listed_origins_by_type' here...
    visit_type = next(iter(listed_origins_by_type))
    assert len(listed_origins_by_type[visit_type]) > NUM_RESULTS

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(origins)

    result = invoke(swh_scheduler, args=("grab-next", visit_type, str(NUM_RESULTS)))
    assert result.exit_code == 0

    out_lines = result.stdout.splitlines()
    assert len(out_lines) == NUM_RESULTS + 1

    fields = out_lines[0].split(",")
    returned_origins = [dict(zip(fields, line.split(","))) for line in out_lines[1:]]

    # Check that we've received origins we had listed in the first place
    assert set(origin["url"] for origin in returned_origins) <= set(
        origin.url for origin in listed_origins_by_type[visit_type]
    )


def test_schedule_next(swh_scheduler, listed_origins_by_type):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    NUM_RESULTS = 10
    # Strict inequality to check that grab_next_visits doesn't return more
    # results than requested
    visit_type = next(iter(listed_origins_by_type))
    assert len(listed_origins_by_type[visit_type]) > NUM_RESULTS

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(origins)

    result = invoke(swh_scheduler, args=("schedule-next", visit_type, str(NUM_RESULTS)))
    assert result.exit_code == 0

    # pull all tasks out of the scheduler
    tasks = swh_scheduler.search_tasks()
    assert len(tasks) == NUM_RESULTS

    scheduled_tasks = {(task.type, task.arguments.kwargs["url"]) for task in tasks}
    all_possible_tasks = {
        (f"load-{origin.visit_type}", origin.url)
        for origin in listed_origins_by_type[visit_type]
    }

    assert scheduled_tasks <= all_possible_tasks


def test_send_to_celery_unknown_visit_type(
    swh_scheduler,
):
    "Calling cli without a known visit type should raise"
    result = invoke(
        swh_scheduler,
        args=("send-origins-from-scheduler-to-celery", "unknown-visit-type"),
        catch_exceptions=True,
    )
    assert "Unknown" in result.output
    assert result.exit_code != 0


@pytest.mark.parametrize(
    "extra_cmd_args",
    [
        [],
        ["--lister-name", "github", "--lister-instance-name", "github"],
        [
            "--absolute-cooldown",
            "24 days",
            "--scheduled-cooldown",
            "24 hours",
            "--failed-cooldown",
            "1 day",
            "--not-found-cooldown",
            "60 days",
        ],
    ],
)
def test_send_origins_from_scheduler_to_celery(
    mocker,
    swh_scheduler,
    task_types,
    swh_scheduler_celery_app,
    listed_origins_by_type,
    extra_cmd_args,
):
    visit_type = next(iter(listed_origins_by_type))

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(origins)

    get_queue_length = mocker.patch(
        "swh.scheduler.celery_backend.config.get_queue_length"
    )
    get_queue_length.return_value = None

    send_task = mocker.patch.object(swh_scheduler_celery_app, "send_task")
    send_task.return_value = None

    cmd_args = ["send-origins-from-scheduler-to-celery", visit_type] + extra_cmd_args

    result = invoke(swh_scheduler, args=tuple(cmd_args))
    assert result.exit_code == 0

    scheduled_tasks = {
        (call[0][0], call[1]["kwargs"]["url"]) for call in send_task.call_args_list
    }

    expected_tasks = {
        (TASK_TYPES[origin.visit_type].backend_name, origin.url)
        for origin in listed_origins_by_type[visit_type]
    }

    assert expected_tasks == scheduled_tasks


def test_update_metrics(swh_scheduler, listed_origins):
    swh_scheduler.record_listed_origins(listed_origins)

    assert swh_scheduler.get_metrics() == []

    result = invoke(swh_scheduler, args=("update-metrics",))

    assert result.exit_code == 0
    assert swh_scheduler.get_metrics() != []


@pytest.mark.parametrize(
    "limit,queue_name_prefix,dry_run,debug",
    [
        (None, "", False, False),
        (10, "", False, False),
        (None, "large-repository", False, False),
        (None, "", True, False),
        (None, "", False, True),
    ],
)
def test_send_origins_from_file_to_celery_cli(
    mocker,
    swh_scheduler,
    swh_scheduler_celery_app,
    task_types,
    listed_origins_by_type,
    limit,
    queue_name_prefix,
    dry_run,
    debug,
):
    visit_type = "test-git"
    origins_to_schedule = listed_origins_by_type[visit_type][:20]

    task_type_param = next(iter(task_types))
    origins_to_send = [o.url for o in origins_to_schedule]

    get_queue_length = mocker.patch(
        "swh.scheduler.celery_backend.config.get_queue_length"
    )
    get_queue_length.return_value = None

    send_task = mocker.patch.object(swh_scheduler_celery_app, "send_task")
    send_task.return_value = None

    extra_cmd_args = []
    if limit:
        extra_cmd_args += ["--limit", limit]
    if queue_name_prefix:
        extra_cmd_args += ["--queue-name-prefix", queue_name_prefix]
    if dry_run:
        extra_cmd_args += ["--dry-run"]
    if debug:
        extra_cmd_args += ["--debug"]

    cmd_args = ["send-origins-from-file-to-celery"] + extra_cmd_args + [task_type_param]

    input_data = ("\n").join(origins_to_send)
    result = invoke(swh_scheduler, args=tuple(cmd_args), input=input_data)
    assert result.exit_code == 0

    if dry_run:
        assert "** DRY-RUN **" in result.output

    if debug:
        assert "Destination queue" in result.output

    actual_scheduled_tasks = {
        (call[1]["name"], call[1]["kwargs"]["url"], call[1]["queue"])
        for call in send_task.call_args_list
    }

    backend_name = task_types[task_type_param].backend_name
    queue = f"{queue_name_prefix}:{backend_name}" if queue_name_prefix else backend_name
    expected_tasks = (
        {}
        if dry_run
        else {
            (backend_name, origin.url, queue)
            for origin in origins_to_schedule[
                : limit if limit else len(origins_to_schedule)
            ]
        }
    )

    assert set(actual_scheduled_tasks) == set(expected_tasks)



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging
import time
from typing import TYPE_CHECKING, List, Tuple

import click
import sentry_sdk

from . import cli

if TYPE_CHECKING:
    from ..interface import SchedulerInterface


@cli.command("start-runner")
@click.option(
    "--period",
    "-p",
    default=0,
    help=(
        "Period (in s) at witch pending tasks are checked and "
        "executed. Set to 0 (default) for a one shot."
    ),
)
@click.option(
    "--task-type",
    "task_type_names",
    multiple=True,
    default=[],
    help=(
        "Task types to schedule. If not provided, this iterates over every "
        "task types referenced in the scheduler backend."
    ),
)
@click.option(
    "--with-priority/--without-priority",
    is_flag=True,
    default=False,
    help=(
        "Determine if those tasks should be the ones with priority or not."
        "By default, this deals with tasks without any priority."
    ),
)
@click.pass_context
def runner(ctx, period, task_type_names, with_priority):
    """Starts a swh-scheduler runner service.

    This process is responsible for checking for ready-to-run tasks and
    schedule them.

    Expected configuration:

    \b
    * :ref:`cli-config-celery`
    * :ref:`cli-config-scheduler`
    """
    from swh.scheduler.celery_backend.config import build_app
    from swh.scheduler.celery_backend.runner import run_ready_tasks

    config = ctx.obj["config"]
    app = build_app(config.get("celery"))
    app.set_current()

    logger = logging.getLogger(__name__ + ".runner")
    scheduler = ctx.obj["scheduler"]
    logger.debug("Scheduler %s", scheduler)
    task_types = []
    for task_type_name in task_type_names:
        task_type = scheduler.get_task_type(task_type_name)
        if not task_type:
            ctx.fail(f"Unknown {task_type_name}")
        task_types.append(task_type)

    try:
        while True:
            logger.debug("Run ready tasks")
            try:
                ntasks = len(run_ready_tasks(scheduler, app, task_types, with_priority))
                if ntasks:
                    logger.info("Scheduled %s tasks", ntasks)
            except Exception:
                logger.exception("Unexpected error in run_ready_tasks()")
                sentry_sdk.capture_exception()
            if not period:
                break
            time.sleep(period)
    except KeyboardInterrupt:
        ctx.exit(0)


@cli.command("start-listener")
@click.pass_context
def listener(ctx):
    """Starts a swh-scheduler listener service.

    This service is responsible for listening at task lifecycle events and
    handle their workflow status in the database.

    Expected configuration:

    \b
    * :ref:`cli-config-celery`
    * :ref:`cli-config-scheduler`
    """
    scheduler_backend = ctx.obj["scheduler"]
    if not scheduler_backend:
        ctx.fail("Scheduler class (local/remote) must be instantiated")

    broker = (
        ctx.obj["config"]
        .get("celery", {})
        .get("task_broker", "amqp://guest@localhost/%2f")
    )

    from swh.scheduler.celery_backend.pika_listener import get_listener

    listener = get_listener(broker, "celeryev.listener", scheduler_backend)
    try:
        listener.start_consuming()
    finally:
        listener.stop_consuming()


@cli.command("schedule-recurrent")
@click.option(
    "--visit-type",
    "visit_types",
    multiple=True,
    default=[],
    help=(
        "Visit types to schedule. If not provided, this iterates over every "
        "corresponding load task types referenced in the scheduler backend."
    ),
)
@click.pass_context
def schedule_recurrent(ctx, visit_types: List[str]):
    """Starts the scheduler for recurrent visits.

    This runs one thread for each visit type, which regularly sends new visits
    to celery.

    Expected configuration:

    \b
    * :ref:`cli-config-celery`
    * :ref:`cli-config-scheduler`
    """
    from queue import Queue

    from swh.scheduler.celery_backend.recurrent_visits import (
        VisitSchedulerThreads,
        logger,
        spawn_visit_scheduler_thread,
        terminate_visit_scheduler_threads,
    )

    config = ctx.obj["config"]
    scheduler = ctx.obj["scheduler"]

    if not visit_types:
        visit_types = []
        # Figure out which visit types exist in the scheduler
        all_task_types = scheduler.get_task_types()
        for task_type in all_task_types:
            if not task_type.type.startswith("load-"):
                # only consider loading tasks as recurring ones, the rest is dismissed
                continue
            # get visit type name from task type
            visit_types.append(task_type.type[5:])
    else:
        # Check that the passed visit types exist in the scheduler
        for visit_type in visit_types:
            task_type_name = f"load-{visit_type}"
            task_type = scheduler.get_task_type(task_type_name)
            if not task_type:
                ctx.fail(f"Unknown task type: {task_type_name}")

    exc_queue: Queue[Tuple[str, BaseException]] = Queue()
    threads: VisitSchedulerThreads = {}

    try:
        # Spawn initial threads
        for visit_type in visit_types:
            spawn_visit_scheduler_thread(threads, exc_queue, config, visit_type)

        # Handle exceptions from child threads
        while True:
            visit_type, exc_info = exc_queue.get(block=True)

            logger.exception(
                "Thread %s died with exception; respawning",
                visit_type,
                exc_info=exc_info,
            )
            sentry_sdk.capture_exception(exc_info)

            dead_thread = threads[visit_type][0]
            dead_thread.join(timeout=1)

            if dead_thread.is_alive():
                logger.warn(
                    "The thread for %s is still alive after sending an exception?! "
                    "Respawning anyway.",
                    visit_type,
                )

            spawn_visit_scheduler_thread(threads, exc_queue, config, visit_type)

    except SystemExit:
        remaining_threads = terminate_visit_scheduler_threads(threads)
        if remaining_threads:
            ctx.exit(1)
        ctx.exit(0)


@cli.command("start-runner-first-visits")
@click.option(
    "--period",
    "-p",
    default=0,
    help=(
        "Period (in s) at witch pending tasks are checked and "
        "executed. Set to 0 (default) for a one shot."
    ),
)
@click.pass_context
def runner_first_visits(ctx, period):
    """Starts a swh-scheduler runner service for first visits.

    This process is responsible for scheduling the first visits for origins registered
    by listers having their first_visits_priority_queue attribute set. It's looping
    forever with a given period.

    """

    from swh.scheduler.celery_backend.first_visits import schedule_first_visits

    logger = logging.getLogger(__name__ + ".runner_first_visits")
    scheduler: SchedulerInterface = ctx.obj["scheduler"]

    try:
        while True:
            logger.debug("Schedule first visits...")
            try:
                nb_first_visits = schedule_first_visits(scheduler)
                logger.info("Scheduled %s first visits", nb_first_visits)
            except Exception:
                logger.exception("Unexpected error in run_high_priority_first_visits()")
                sentry_sdk.capture_exception()
            if not period:
                break
            time.sleep(period)
    except KeyboardInterrupt:
        ctx.exit(0)


@cli.command("rpc-serve")
@click.option("--host", default="0.0.0.0", help="Host to run the scheduler server api")
@click.option("--port", default=5008, type=click.INT, help="Binding port of the server")
@click.option(
    "--debug/--nodebug",
    default=None,
    help=(
        "Indicates if the server should run in debug mode. "
        "Defaults to True if log-level is DEBUG, False otherwise."
    ),
)
@click.pass_context
def rpc_server(ctx, host, port, debug):
    """Starts a swh-scheduler API HTTP server.

    Expected configuration:

    \b
    * :ref:`cli-config-celery`
    * :ref:`cli-config-scheduler`
    """
    if ctx.obj["config"]["scheduler"]["cls"] == "remote":
        click.echo(
            "The API server can only be started with a 'local' " "configuration",
            err=True,
        )
        ctx.exit(1)

    from swh.scheduler.api import server

    server.app.config.update(ctx.obj["config"])
    if debug is None:
        debug = ctx.obj["log_level"] <= logging.DEBUG
    server.app.run(host, port=port, debug=bool(debug))



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import functools
from itertools import permutations
from typing import Any, Dict, List
from unittest.mock import Mock

import attr
import pytest

from swh.model.hashutil import hash_to_bytes
from swh.scheduler.journal_client import (
    from_position_offset_to_days,
    max_date,
    next_visit_queue_position,
    process_journal_objects,
)
from swh.scheduler.model import LastVisitStatus, ListedOrigin, OriginVisitStats
from swh.scheduler.utils import utcnow


def test_journal_client_origin_visit_status_from_journal_fail(swh_scheduler):
    process_fn = functools.partial(
        process_journal_objects,
        scheduler=swh_scheduler,
    )

    with pytest.raises(AssertionError, match="Got unexpected origin_visit"):
        process_fn(
            {
                "origin_visit": [
                    {"url": "http://foobar.baz"},
                ]
            }
        )

    with pytest.raises(AssertionError, match="Expected origin_visit_status"):
        process_fn({})


ONE_DAY = datetime.timedelta(days=1)
ONE_YEAR = datetime.timedelta(days=366)

DATE3 = utcnow()
DATE2 = DATE3 - ONE_DAY
DATE1 = DATE2 - ONE_DAY


assert DATE1 < DATE2 < DATE3


@pytest.mark.parametrize(
    "dates,expected_max_date",
    [
        ((DATE1,), DATE1),
        ((None, DATE2), DATE2),
        ((DATE1, None), DATE1),
        ((DATE1, DATE2), DATE2),
        ((DATE2, DATE1), DATE2),
        ((DATE1, DATE2, DATE3), DATE3),
        ((None, DATE2, DATE3), DATE3),
        ((None, None, DATE3), DATE3),
        ((DATE1, None, DATE3), DATE3),
    ],
)
def test_max_date(dates, expected_max_date):
    assert max_date(*dates) == expected_max_date


def test_max_date_raise():
    with pytest.raises(ValueError, match="valid datetime"):
        max_date()
    with pytest.raises(ValueError, match="valid datetime"):
        max_date(None)
    with pytest.raises(ValueError, match="valid datetime"):
        max_date(None, None)


def test_journal_client_origin_visit_status_from_journal_ignored_status(swh_scheduler):
    """Only final statuses (full, partial) are important, the rest remain ignored."""
    # Trace method calls on the swh_scheduler
    swh_scheduler = Mock(wraps=swh_scheduler)

    visit_statuses = [
        {
            "origin": "foo",
            "visit": 1,
            "status": "created",
            "date": utcnow(),
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 1,
            "status": "ongoing",
            "date": utcnow(),
            "type": "svn",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    # All messages have been ignored: no stats have been upserted
    swh_scheduler.origin_visit_stats_upsert.assert_not_called()


def test_journal_client_ignore_missing_type(swh_scheduler):
    """Ignore statuses with missing type key"""
    # Trace method calls on the swh_scheduler
    swh_scheduler = Mock(wraps=swh_scheduler)

    date = utcnow()
    snapshot = hash_to_bytes("dddcc0710eb6cf9efd5b920a8453e1e07157bddd")
    visit_statuses = [
        {
            "origin": "foo",
            "visit": 1,
            "status": "full",
            "date": date,
            "snapshot": snapshot,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    # The message has been ignored: no stats have been upserted
    swh_scheduler.origin_visit_stats_upsert.assert_not_called()


def assert_visit_stats_ok(
    actual_visit_stats: OriginVisitStats,
    expected_visit_stats: OriginVisitStats,
    ignore_fields: List[str] = ["next_visit_queue_position"],
):
    """Utility test function to ensure visits stats read from the backend are in the right
    shape. The comparison on the next_visit_queue_position will be dealt with in
    dedicated tests so it's not tested in tests that are calling this function.

    """
    fields = attr.fields_dict(OriginVisitStats)
    defaults: Dict[str, Any] = {field: fields[field].default for field in ignore_fields}

    actual_visit_stats = attr.evolve(actual_visit_stats, **defaults)
    assert actual_visit_stats == expected_visit_stats


def test_journal_client_origin_visit_status_from_journal_last_not_found(swh_scheduler):
    visit_status = {
        "origin": "foo",
        "visit": 1,
        "status": "not_found",
        "date": DATE1,
        "type": "git",
        "snapshot": None,
    }

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_visit=visit_status["date"],
            last_visit_status=LastVisitStatus.not_found,
            next_position_offset=4,
            successive_visits=1,
        ),
    )

    visit_statuses = [
        {
            "origin": "foo",
            "visit": 3,
            "status": "not_found",
            "date": DATE2,
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "foo",
            "visit": 4,
            "status": "not_found",
            "date": DATE3,
            "type": "git",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.not_found,
            next_position_offset=6,
            successive_visits=3,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_failed(swh_scheduler):
    visit_statuses = [
        {
            "origin": "foo",
            "visit": 1,
            "status": "partial",
            "date": utcnow(),
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 1,
            "status": "full",
            "date": DATE1,
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 2,
            "status": "full",
            "date": DATE2,
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "full",
            "date": DATE3,
            "type": "git",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("bar", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="bar",
            visit_type="git",
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.failed,
            next_position_offset=6,
            successive_visits=3,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_failed2(swh_scheduler):
    visit_statuses = [
        {
            "origin": "bar",
            "visit": 2,
            "status": "failed",
            "date": DATE1,
            "type": "git",
            "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "failed",
            "date": DATE2,
            "type": "git",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("bar", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="bar",
            visit_type="git",
            last_visit=DATE2,
            last_visit_status=LastVisitStatus.failed,
            next_position_offset=5,
            successive_visits=2,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_successful(swh_scheduler):
    visit_statuses = [
        {
            "origin": "bar",
            "visit": 1,
            "status": "partial",
            "date": utcnow(),
            "type": "git",
            "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        },
        {
            "origin": "foo",
            "visit": 1,
            "status": "full",
            "date": DATE1,
            "type": "git",
            "snapshot": hash_to_bytes("eeecc0710eb6cf9efd5b920a8453e1e07157bfff"),
        },
        {
            "origin": "foo",
            "visit": 2,
            "status": "partial",
            "date": DATE2,
            "type": "git",
            "snapshot": hash_to_bytes("aaacc0710eb6cf9efd5b920a8453e1e07157baaa"),
        },
        {
            "origin": "foo",
            "visit": 3,
            "status": "full",
            "date": DATE3,
            "type": "git",
            "snapshot": hash_to_bytes("dddcc0710eb6cf9efd5b920a8453e1e07157bddd"),
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE3,
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("dddcc0710eb6cf9efd5b920a8453e1e07157bddd"),
            next_position_offset=0,
            successive_visits=3,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_uneventful(swh_scheduler):
    visit_status = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE3 + ONE_DAY,
        "type": "git",
        "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
    }

    # Let's insert some visit stats with some previous visit information
    swh_scheduler.origin_visit_stats_upsert(
        [
            OriginVisitStats(
                url=visit_status["origin"],
                visit_type=visit_status["type"],
                last_successful=DATE2,
                last_visit=DATE3,
                last_visit_status=LastVisitStatus.failed,
                last_snapshot=visit_status["snapshot"],
                next_visit_queue_position=None,
                next_position_offset=4,
                successive_visits=1,
            )
        ]
    )

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get(
        [(visit_status["origin"], visit_status["type"])]
    )

    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url=visit_status["origin"],
            visit_type=visit_status["type"],
            last_visit=DATE3 + ONE_DAY,
            last_successful=DATE3 + ONE_DAY,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=visit_status["snapshot"],
            next_visit_queue_position=None,
            next_position_offset=5,
            successive_visits=1,
        ),
    )


VISIT_STATUSES = [
    {**ovs, "date": DATE1 + n * ONE_DAY}
    for n, ovs in enumerate(
        [
            {
                "origin": "foo",
                "type": "git",
                "visit": 1,
                "status": "created",
                "snapshot": None,
            },
            {
                "origin": "foo",
                "type": "git",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "foo",
                "type": "git",
                "visit": 2,
                "status": "created",
                "snapshot": None,
            },
            {
                "origin": "foo",
                "type": "git",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
        ]
    )
]


@pytest.mark.parametrize(
    "visit_statuses", permutations(VISIT_STATUSES, len(VISIT_STATUSES))
)
def test_journal_client_origin_visit_status_permutation0(visit_statuses, swh_scheduler):
    """Ensure out of order topic subscription ends up in the same final state"""
    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    visit_stats = actual_origin_visit_stats[0]
    assert_visit_stats_ok(
        visit_stats,
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE1 + 3 * ONE_DAY,
            last_visit=DATE1 + 3 * ONE_DAY,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        ),
        ignore_fields=[
            "next_visit_queue_position",
            "next_position_offset",
            "successive_visits",
        ],
    )

    # We ignore out of order messages, so the next_position_offset isn't exact
    # depending on the permutation. What matters is consistency of the final
    # dates (last_visit and last_successful).
    assert 4 <= visit_stats.next_position_offset <= 5
    # same goes for successive_visits
    assert 1 <= visit_stats.successive_visits <= 2


VISIT_STATUSES_1 = [
    {**ovs, "date": DATE1 + n * ONE_DAY}
    for n, ovs in enumerate(
        [
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 1,
                "status": "partial",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 3,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 4,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
        ]
    )
]


@pytest.mark.parametrize(
    "visit_statuses", permutations(VISIT_STATUSES_1, len(VISIT_STATUSES_1))
)
def test_journal_client_origin_visit_status_permutation1(visit_statuses, swh_scheduler):
    """Ensure out of order topic subscription ends up in the same final state"""
    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_visit_stats = swh_scheduler.origin_visit_stats_get([("cavabarder", "hg")])
    visit_stats = actual_visit_stats[0]

    assert_visit_stats_ok(
        visit_stats,
        OriginVisitStats(
            url="cavabarder",
            visit_type="hg",
            last_successful=DATE1 + 3 * ONE_DAY,
            last_visit=DATE1 + 3 * ONE_DAY,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
        ),
        ignore_fields=[
            "next_visit_queue_position",
            "next_position_offset",
            "successive_visits",
        ],
    )

    # We ignore out of order messages, so the next_position_offset isn't exact
    # depending on the permutation. What matters is consistency of the final
    # dates (last_visit and last_successful).
    assert 2 <= visit_stats.next_position_offset <= 5
    # same goes for successive_visits
    assert 1 <= visit_stats.successive_visits <= 4


VISIT_STATUSES_2 = [
    {**ovs, "date": DATE1 + n * ONE_DAY}
    for n, ovs in enumerate(
        [
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("0000000000000000000000000000000000000000"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("1111111111111111111111111111111111111111"),
            },
            {
                "origin": "iciaussi",
                "type": "hg",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("2222222222222222222222222222222222222222"),
            },
            {
                "origin": "iciaussi",
                "type": "hg",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("3333333333333333333333333333333333333333"),
            },
            {
                "origin": "cavabarder",
                "type": "git",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("4444444444444444444444444444444444444444"),
            },
            {
                "origin": "cavabarder",
                "type": "git",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("5555555555555555555555555555555555555555"),
            },
            {
                "origin": "iciaussi",
                "type": "git",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("6666666666666666666666666666666666666666"),
            },
            {
                "origin": "iciaussi",
                "type": "git",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("7777777777777777777777777777777777777777"),
            },
        ]
    )
]


def test_journal_client_origin_visit_status_after_grab_next_visits(
    swh_scheduler, stored_lister
):
    """Ensure OriginVisitStat entries created in the db as a result of calling
    grab_next_visits() do not mess the OriginVisitStats upsert mechanism.

    """

    listed_origins = [
        ListedOrigin(lister_id=stored_lister.id, url=url, visit_type=visit_type)
        for (url, visit_type) in set((v["origin"], v["type"]) for v in VISIT_STATUSES_2)
    ]
    swh_scheduler.record_listed_origins(listed_origins)
    before = utcnow()
    swh_scheduler.grab_next_visits(
        visit_type="git", count=10, policy="oldest_scheduled_first"
    )
    after = utcnow()

    assert swh_scheduler.origin_visit_stats_get([("cavabarder", "hg")]) == []
    assert swh_scheduler.origin_visit_stats_get([("cavabarder", "git")])[0] is not None

    process_journal_objects(
        {"origin_visit_status": VISIT_STATUSES_2}, scheduler=swh_scheduler
    )

    for url in ("cavabarder", "iciaussi"):
        ovs = swh_scheduler.origin_visit_stats_get([(url, "git")])[0]
        assert before <= ovs.last_scheduled <= after

        ovs = swh_scheduler.origin_visit_stats_get([(url, "hg")])[0]
        assert ovs.last_scheduled is None

    ovs = swh_scheduler.origin_visit_stats_get([("cavabarder", "git")])[0]
    assert ovs.last_successful == DATE1 + 5 * ONE_DAY
    assert ovs.last_visit == DATE1 + 5 * ONE_DAY
    assert ovs.last_visit_status == LastVisitStatus.successful
    assert ovs.last_snapshot == hash_to_bytes(
        "5555555555555555555555555555555555555555"
    )


def test_journal_client_origin_visit_status_duplicated_messages(swh_scheduler):
    """A duplicated message must be ignored"""
    visit_status = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE1,
        "type": "git",
        "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
    }

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE1,
            last_visit=DATE1,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            successive_visits=1,
        ),
    )


def test_journal_client_origin_visit_status_several_upsert(swh_scheduler):
    """An old message updates old information"""
    visit_status1 = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE1,
        "type": "git",
        "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
    }

    visit_status2 = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE2,
        "type": "git",
        "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
    }

    process_journal_objects(
        {"origin_visit_status": [visit_status2]}, scheduler=swh_scheduler
    )

    process_journal_objects(
        {"origin_visit_status": [visit_status1]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE2,
            last_visit=DATE2,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            next_position_offset=4,
            successive_visits=1,
        ),
    )


VISIT_STATUSES_SAME_SNAPSHOT = [
    {**ovs, "date": DATE1 + n * ONE_YEAR}
    for n, ovs in enumerate(
        [
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 3,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 4,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 4,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
        ]
    )
]


@pytest.mark.parametrize(
    "visit_statuses",
    permutations(VISIT_STATUSES_SAME_SNAPSHOT, len(VISIT_STATUSES_SAME_SNAPSHOT)),
)
def test_journal_client_origin_visit_statuses_same_snapshot_permutation(
    visit_statuses, swh_scheduler
):
    """Ensure out of order topic subscription ends up in the same final state"""
    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get(
        [("cavabarder", "hg")]
    )
    visit_stats = actual_origin_visit_stats[0]
    assert_visit_stats_ok(
        visit_stats,
        OriginVisitStats(
            url="cavabarder",
            visit_type="hg",
            last_successful=DATE1 + 2 * ONE_YEAR,
            last_visit=DATE1 + 2 * ONE_YEAR,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
        ),
        ignore_fields=[
            "next_visit_queue_position",
            "next_position_offset",
            "successive_visits",
        ],
    )

    # We ignore out of order messages, so the next_position_offset isn't exact
    # depending on the permutation. What matters is consistency of the final
    # dates (last_visit and last_successful).
    assert 4 <= visit_stats.next_position_offset <= 6
    # same goes for successive_visits
    assert 1 <= visit_stats.successive_visits <= 3


@pytest.mark.parametrize(
    "position_offset, interval",
    [
        (0, 1),
        (1, 1),
        (2, 2),
        (3, 2),
        (4, 2),
        (5, 4),
        (6, 16),
        (7, 64),
        (8, 256),
        (9, 1024),
        (10, 4096),
        (11, 16384),
    ],
)
def test_journal_client_from_position_offset_to_days(position_offset, interval):
    assert from_position_offset_to_days(position_offset) == interval


def test_journal_client_from_position_offset_to_days_only_positive_input():
    with pytest.raises(AssertionError):
        from_position_offset_to_days(-1)


@pytest.mark.parametrize(
    "fudge_factor,next_position_offset",
    [
        (0.01, 1),
        (-0.01, 5),
        (0.1, 8),
        (-0.1, 10),
    ],
)
def test_next_visit_queue_position(mocker, fudge_factor, next_position_offset):
    mock_random = mocker.patch("swh.scheduler.journal_client.random.uniform")
    mock_random.return_value = fudge_factor

    actual_position = next_visit_queue_position(
        {},
        {
            "next_position_offset": next_position_offset,
            "visit_type": "svn",
        },
    )

    assert actual_position == int(
        24
        * 3600
        * from_position_offset_to_days(next_position_offset)
        * (1 + fudge_factor)
    )

    assert mock_random.called


@pytest.mark.parametrize(
    "fudge_factor,next_position_offset",
    [
        (0.02, 2),
        (-0.02, 3),
        (0, 7),
        (-0.09, 9),
    ],
)
def test_next_visit_queue_position_with_state(
    mocker, fudge_factor, next_position_offset
):
    mock_random = mocker.patch("swh.scheduler.journal_client.random.uniform")
    mock_random.return_value = fudge_factor

    actual_position = next_visit_queue_position(
        {"git": 0},
        {
            "next_position_offset": next_position_offset,
            "visit_type": "git",
        },
    )

    assert actual_position == int(
        24
        * 3600
        * from_position_offset_to_days(next_position_offset)
        * (1 + fudge_factor)
    )

    assert mock_random.called


@pytest.mark.parametrize(
    "fudge_factor,next_position_offset",
    [
        (0.03, 3),
        (-0.03, 4),
        (0.08, 7),
        (-0.08, 9),
    ],
)
def test_next_visit_queue_position_with_next_visit_queue(
    mocker, fudge_factor, next_position_offset
):
    mock_random = mocker.patch("swh.scheduler.journal_client.random.uniform")
    mock_random.return_value = fudge_factor

    actual_position = next_visit_queue_position(
        {},
        {
            "next_position_offset": next_position_offset,
            "visit_type": "hg",
            "next_visit_queue_position": 0,
        },
    )

    assert actual_position == int(
        24
        * 3600
        * from_position_offset_to_days(next_position_offset)
        * (1 + fudge_factor)
    )

    assert mock_random.called


def test_disable_failing_origins(swh_scheduler):
    """Origin with too many failed attempts ends up being deactivated in the scheduler."""

    # actually store the origin in the scheduler so we can check it's deactivated in the
    # end.
    lister = swh_scheduler.get_or_create_lister(
        name="something", instance_name="something"
    )
    origin = ListedOrigin(
        url="bar", enabled=True, visit_type="svn", lister_id=lister.id
    )
    swh_scheduler.record_listed_origins([origin])

    visit_statuses = [
        {
            "origin": "bar",
            "visit": 2,
            "status": "failed",
            "date": DATE1,
            "type": "svn",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "failed",
            "date": DATE2,
            "type": "svn",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "failed",
            "date": DATE3,
            "type": "svn",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("bar", "svn")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="bar",
            visit_type="svn",
            last_successful=None,
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.failed,
            next_position_offset=6,
            successive_visits=3,
        ),
    )

    # Now check that the origin in question is disabled
    actual_page = swh_scheduler.get_listed_origins(url="bar", enabled=False)

    assert len(actual_page.results) == 1
    assert actual_page.next_page_token is None

    for origin in actual_page.results:
        assert origin.enabled is False
        assert origin.lister_id == lister.id
        assert origin.url == "bar"
        assert origin.visit_type == "svn"



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import cli
from ..utils import create_origin_tasks

if TYPE_CHECKING:
    from typing import Dict, Iterable, List, Optional
    from uuid import UUID

    from ..interface import SchedulerInterface, TaskPolicy
    from ..model import ListedOrigin


@cli.group("origin")
@click.pass_context
def origin(ctx):
    """Manipulate listed origins."""
    if not ctx.obj["scheduler"]:
        ctx.fail("Scheduler class (local/remote) must be instantiated")


def format_origins(
    ctx: click.Context,
    origins: List[ListedOrigin],
    fields: Optional[List[str]] = None,
    with_header: bool = True,
) -> Iterable[str]:
    """Format a list of origins as CSV.

    Arguments:
       origins: list of origins to output
       fields: optional list of fields to output (defaults to all fields)
       with_header: if True, output a CSV header.
    """
    import csv
    from io import StringIO

    import attr

    from ..model import ListedOrigin

    expected_fields = [field.name for field in attr.fields(ListedOrigin)]
    if not fields:
        fields = expected_fields

    unknown_fields = set(fields) - set(expected_fields)
    if unknown_fields:
        ctx.fail("Unknown ListedOrigin field(s): %s" % ", ".join(unknown_fields))

    output = StringIO()
    writer = csv.writer(output)

    def csv_row(data):
        """Return a single CSV-formatted row. We clear the output buffer after we're
        done to keep it reasonably sized."""
        writer.writerow(data)
        output.seek(0)
        ret = output.read().rstrip()
        output.seek(0)
        output.truncate()
        return ret

    if with_header:
        yield csv_row(fields)

    for origin in origins:
        yield csv_row(str(getattr(origin, field)) for field in fields)


@origin.command("grab-next")
@click.option(
    "--policy", "-p", default="oldest_scheduled_first", help="Scheduling policy"
)
@click.option(
    "--fields", "-f", default=None, help="Listed origin fields to print on output"
)
@click.option(
    "--with-header/--without-header",
    is_flag=True,
    default=True,
    help="Print the CSV header?",
)
@click.argument("type", type=str)
@click.argument("count", type=int)
@click.pass_context
def grab_next(
    ctx,
    policy: TaskPolicy,
    fields: Optional[str],
    with_header: bool,
    type: str,
    count: int,
):
    """Grab the next COUNT origins to visit using the TYPE loader from the
    listed origins table."""

    if fields:
        parsed_fields: Optional[List[str]] = fields.split(",")
    else:
        parsed_fields = None

    scheduler = ctx.obj["scheduler"]

    origins = scheduler.grab_next_visits(type, count, policy=policy)
    for line in format_origins(
        ctx, origins, fields=parsed_fields, with_header=with_header
    ):
        click.echo(line)


@origin.command("schedule-next")
@click.option(
    "--policy", "-p", default="oldest_scheduled_first", help="Scheduling policy"
)
@click.argument("type", type=str)
@click.argument("count", type=int)
@click.pass_context
def schedule_next(ctx, policy: TaskPolicy, type: str, count: int):
    """Send the next COUNT origin visits of the TYPE loader to the scheduler as
    one-shot tasks."""
    from ..utils import utcnow
    from .utils import pretty_print_task

    scheduler = ctx.obj["scheduler"]

    origins = scheduler.grab_next_visits(type, count, policy=policy)

    created = scheduler.create_tasks(
        [
            task.evolve(
                policy="oneshot",
                next_run=utcnow(),
                retries_left=1,
            )
            for task in create_origin_tasks(origins, scheduler)
        ]
    )

    output = ["Created %d tasks\n" % len(created)]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo_via_pager("\n".join(output))


@origin.command("send-origins-from-scheduler-to-celery")
@click.option(
    "--policy", "-p", default="oldest_scheduled_first", help="Scheduling policy"
)
@click.option(
    "--queue",
    "-q",
    help="Target celery queue",
    type=str,
)
@click.option(
    "--tablesample",
    help="Table sampling percentage",
    type=float,
)
@click.option(
    "--only-enabled/--only-disabled",
    "enabled",
    is_flag=True,
    default=True,
    help="""Determine whether we want to scheduled enabled or disabled origins. As default, we
            want to reasonably deal with enabled origins. For some edge case though, we
            might want the disabled ones.""",
)
@click.option(
    "--lister-name",
    default=None,
    help="Limit origins to those listed from lister with provided name",
)
@click.option(
    "--lister-instance-name",
    default=None,
    help="Limit origins to those listed from lister with instance name",
)
@click.option(
    "--absolute-cooldown",
    "absolute_cooldown_str",
    default="12 hours",
    help="Minimal interval between two visits of the same origin",
)
@click.option(
    "--scheduled-cooldown",
    "scheduled_cooldown_str",
    default="7 days",
    help="Minimal interval to wait before scheduling the same origins again",
)
@click.option(
    "--not-found-cooldown",
    "not_found_cooldown_str",
    default="14 days",
    help="The minimal interval to wait before rescheduling not_found origins",
)
@click.option(
    "--failed-cooldown",
    "failed_cooldown_str",
    default="31 days",
    help="Minimal interval to wait before rescheduling failed origins",
)
@click.argument("visit_type_name", type=str)
@click.pass_context
def send_from_scheduler_to_celery_cli(
    ctx,
    policy: str,
    queue: Optional[str],
    tablesample: Optional[float],
    visit_type_name: str,
    enabled: bool,
    lister_name: Optional[str] = None,
    lister_instance_name: Optional[str] = None,
    absolute_cooldown_str: Optional[str] = None,
    scheduled_cooldown_str: Optional[str] = None,
    failed_cooldown_str: Optional[str] = None,
    not_found_cooldown_str: Optional[str] = None,
):
    """Send next origin visits of VISIT_TYPE_NAME to celery, filling the queue."""

    from swh.scheduler.celery_backend.utils import get_loader_task_type, send_to_celery

    from .utils import parse_time_interval

    absolute_cooldown = (
        parse_time_interval(absolute_cooldown_str) if absolute_cooldown_str else None
    )
    scheduled_cooldown = (
        parse_time_interval(scheduled_cooldown_str) if scheduled_cooldown_str else None
    )
    failed_cooldown = (
        parse_time_interval(failed_cooldown_str) if failed_cooldown_str else None
    )
    not_found_cooldown = (
        parse_time_interval(not_found_cooldown_str) if not_found_cooldown_str else None
    )

    scheduler = ctx.obj["scheduler"]

    task_type = get_loader_task_type(scheduler, visit_type_name)
    if not task_type:
        ctx.fail(f"Unknown task type {task_type}.")
        assert False  # for mypy

    queue_name = queue or task_type.backend_name

    send_to_celery(
        scheduler,
        visit_type_to_queue={visit_type_name: queue_name},
        policy=policy,
        tablesample=tablesample,
        enabled=enabled,
        lister_name=lister_name,
        lister_instance_name=lister_instance_name,
        absolute_cooldown=absolute_cooldown,
        scheduled_cooldown=scheduled_cooldown,
        failed_cooldown=failed_cooldown,
        not_found_cooldown=not_found_cooldown,
    )


@origin.command("update-metrics")
@click.option("--lister", default=None, help="Only update metrics for this lister")
@click.option(
    "--instance", default=None, help="Only update metrics for this lister instance"
)
@click.pass_context
def update_metrics(ctx, lister: Optional[str], instance: Optional[str]):
    """Update the scheduler metrics on listed origins.

    Examples:
       swh scheduler origin update-metrics
       swh scheduler origin update-metrics --lister github
       swh scheduler origin update-metrics --lister phabricator --instance llvm
    """
    import json

    scheduler: SchedulerInterface = ctx.obj["scheduler"]

    lister_id: Optional[UUID] = None
    if lister is not None:
        lister_instance = scheduler.get_lister(name=lister, instance_name=instance)
        if not lister_instance:
            click.echo(f"Lister not found: {lister} instance={instance}")
            ctx.exit(2)
            assert False  # for mypy

        lister_id = lister_instance.id

    def dictify_metrics(d):
        return {k: str(v) for (k, v) in d.to_dict().items()}

    ret = scheduler.update_metrics(lister_id=lister_id)
    click.echo(json.dumps(list(map(dictify_metrics, ret)), indent=4, sort_keys=True))


@origin.command("check-listed-origins")
@click.option(
    "--list",
    "-l",
    is_flag=True,
    help="Display listed origins (disabled by default).",
    required=False,
)
@click.argument("lister_name", nargs=1, required=True)
@click.argument("instance_name", nargs=1, required=True)
@click.pass_context
def check_listed_origins_cli(ctx, list, lister_name, instance_name):
    """
    Check listed origins registered in the scheduler database.
    """

    from tabulate import tabulate

    from .utils import check_listed_origins

    scheduler = ctx.obj["scheduler"]
    listed_origins = check_listed_origins(
        scheduler=scheduler,
        lister_name=lister_name,
        instance_name=instance_name,
    )
    listed_origins_table = [
        (origin.url, str(origin.last_seen), str(origin.last_update))
        for origin in listed_origins
    ]
    headers = ["url", "last_seen", "last_update"]

    if list:
        print(tabulate(listed_origins_table, headers))
    print(
        f"\nForge {instance_name} ({lister_name}) has {len(listed_origins)} \
listed origins in the scheduler database."
    )


def _print_status_summary(
    lister_name: str, instance_name: str, status_counters: Dict, watch: bool = False
) -> None:
    """Print a status of the ingestion and a small success rate summary."""
    if watch:
        suffix_str = " ingestion is still in progress..."
    else:
        suffix_str = f" ({lister_name}) has {status_counters['total']} \
scheduled ingests in the scheduler."

    print(f"\nForge {instance_name}{suffix_str}")
    for status, counter in status_counters.items():
        print(f"{status:<12}: {counter}")

    success_rate = status_counters["successful"] / status_counters["total"] * 100
    print(f"{'success rate':<12}: {success_rate:.2f}%")


@origin.command("check-ingested-origins")
@click.option(
    "--list",
    "-l",
    "with_listing",
    is_flag=True,
    help="Display listed origins (disabled by default).",
    required=False,
)
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    required=False,
    help="Watch periodically for ingestion status.",
)
@click.option(
    "--watch-period",
    required=False,
    default="30 minutes",
    help="Watch period ingestion.",
)
@click.argument("lister_name", nargs=1, required=True)
@click.argument("instance_name", nargs=1, required=True)
@click.pass_context
def check_ingested_origins_cli(
    ctx, with_listing, watch, watch_period, lister_name, instance_name
):
    """
    Check the origins marked as ingested in the scheduler database.
    """

    from time import sleep

    from .utils import check_listed_origins, count_ingested_origins, parse_time_interval

    scheduler = ctx.obj["scheduler"]
    listed_origins = check_listed_origins(
        scheduler=scheduler,
        lister_name=lister_name,
        instance_name=instance_name,
    )
    ids = [(origin.url, origin.visit_type) for origin in listed_origins]
    status_counters, _ = count_ingested_origins(
        scheduler=scheduler,
        ids=ids,
        instance_name=instance_name,
    )

    if watch:
        watch_period_seconds = parse_time_interval(watch_period).total_seconds()
        while status_counters["None"] != 0:
            status_counters, _ = count_ingested_origins(
                scheduler=scheduler,
                ids=ids,
                instance_name=instance_name,
            )
            if status_counters["None"] > 0:
                _print_status_summary(
                    lister_name, instance_name, status_counters, watch
                )

            sleep(watch_period_seconds)

    status_counters, ingested_origins_table = count_ingested_origins(
        scheduler=scheduler,
        ids=ids,
        instance_name=instance_name,
        with_listing=with_listing,
    )

    if with_listing:
        from tabulate import tabulate

        headers = ("url", "last_visit_status", "last_visit")
        print(tabulate(ingested_origins_table, headers))

    _print_status_summary(lister_name, instance_name, status_counters)


@origin.command("send-origins-from-file-to-celery")
@click.option(
    "--queue-name-prefix",
    help=(
        "Prefix to add to the default queue name (if needed). Usually needed to treat "
        "special origins (e.g. large repositories, fill-in-the-hole datasets, ... ) "
        "in another dedicated queue."
    ),
)
@click.option(
    "--threshold",
    help="Threshold override for the queue.",
    type=click.INT,
    default=1000,
)
@click.option(
    "--limit",
    help="Number of origins to send. Usually to limit to a small number for debug purposes.",
    type=click.INT,
    default=None,
)
@click.option(
    "--waiting-period",
    help="Waiting time between checks",
    type=click.INT,
    default=10,
)
@click.option("--dry-run", is_flag=True, help="Print only messages to send to celery")
@click.option("--debug", is_flag=True, default=False, help="Print extra messages")
# Scheduler task type (e.g. load-git, load-svn, ...)
@click.argument(
    "task_type",
    nargs=1,
    required=True,
)
# Dataset file of origins to schedule, use '-' when piping to the cli.
@click.argument(
    "file_input",
    type=click.File("r"),
    required=False,
    default="-",
)
# Extra options passed directly to the task
@click.argument(
    "options",
    nargs=-1,
    required=False,
)
@click.pass_context
def send_origins_from_file_to_celery(
    ctx,
    queue_name_prefix,
    threshold,
    waiting_period,
    dry_run,
    task_type,
    options,
    debug,
    limit,
    file_input,
):
    """Send origins directly from file/stdin to celery, filling the queue according to
    its standard configuration (and some optional adjustments).

    Arguments:

        TASK_TYPE: Scheduler task type (e.g. load-git, load-svn, ...)

        INPUT: Dataset file of origins to schedule, use '-' when piping to the cli.

        OPTIONS: Extra options (in the key=value form, e.g. base_git_url=<foo>) passed
                 directly to the task to be scheduled

    """
    from functools import partial
    from time import sleep

    from swh.scheduler.celery_backend.config import app, get_available_slots
    from swh.scheduler.cli.utils import parse_options

    from .origin_utils import (
        TASK_ARGS_GENERATOR_CALLABLES,
        get_scheduler_task_type,
        lines_to_task_args,
    )

    scheduler = ctx.obj["scheduler"]

    try:
        task_type_info = get_scheduler_task_type(scheduler, task_type)
    except ValueError as e:
        ctx.fail(e)

    if debug:
        click.echo(f"Scheduler task information: {task_type_info}")

    celery_task_name = task_type_info.backend_name
    if queue_name_prefix:
        queue_name = f"{queue_name_prefix}:{celery_task_name}"
    else:
        queue_name = celery_task_name

    if debug:
        click.echo(f"Destination queue: {queue_name}")

    if not threshold:
        threshold = task_type_info.max_queue_length

    # Compute the callable function to generate the tasks to schedule. Tasks are read
    # out of the file_input.
    task_fn = partial(
        TASK_ARGS_GENERATOR_CALLABLES.get(
            task_type, partial(lines_to_task_args, columns=["url"])
        ),
        lines=file_input,
    )

    (extra_args, extra_kwargs) = parse_options(options)

    if debug:
        print(f"Task extra_args: {extra_args}")
        print(f"task Extra_kwargs: {extra_kwargs}")

    limit_reached = False
    while True:
        throttled = False
        remains_data = False
        pending_tasks = []

        if limit_reached:
            break

        # we can send new tasks, compute how many we can send
        nb_tasks_to_send = get_available_slots(app, queue_name, threshold)

        if debug:
            click.echo(f"Nb tasks to send : {nb_tasks_to_send}")

        if nb_tasks_to_send > 0:
            count = 0
            for _task in task_fn(**extra_kwargs):
                pending_tasks.append(_task)
                count += 1

                if debug:
                    click.echo(f"Task: {_task}")

                if limit is not None and count >= limit:
                    click.echo(f"Limit {limit} messages specified reached, stopping.")
                    limit_reached = True
                    break

                if count >= nb_tasks_to_send:
                    throttled = True
                    remains_data = True
                    break

            if not pending_tasks:
                # check for some more data on stdin
                if not remains_data:
                    # if no more data, we break to exit
                    break

            from kombu.utils.uuid import uuid

            for _task_args in pending_tasks:
                send_task_kwargs = dict(
                    name=celery_task_name,
                    task_id=uuid(),
                    args=tuple(_task_args["args"]) + tuple(extra_args),
                    kwargs=_task_args["kwargs"],
                    queue=queue_name,
                )
                if dry_run:
                    click.echo(
                        f"** DRY-RUN ** Call app.send_task with: {send_task_kwargs}"
                    )
                else:
                    app.send_task(**send_task_kwargs)
                    click.echo(send_task_kwargs)

        else:
            throttled = True

        if throttled:
            sleep(waiting_period)



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from swh.scheduler.interface import SchedulerInterface


logger = logging.getLogger(__name__)


def schedule_first_visits(backend: SchedulerInterface):
    """Schedule first visits with high priority for origins registered by listers
    having the first_visits_priority_queue attribute set.

    """
    from functools import partial

    from swh.core.api.classes import stream_results
    from swh.core.utils import grouper
    from swh.scheduler.utils import utcnow

    from .utils import get_loader_task_type, send_to_celery

    nb_first_visits = 0
    for lister in backend.get_listers(with_first_visits_to_schedule=True):
        visit_types = backend.get_visit_types_for_listed_origins(lister)
        visit_type_to_queue = {}

        for visit_type in visit_types:
            task_type = get_loader_task_type(backend, visit_type)
            if not task_type:
                raise ValueError(f"Unknown task type for visit type {visit_type}.")
            visit_type_to_queue[visit_type] = (
                f"{lister.first_visits_queue_prefix}:{task_type.backend_name}"
            )

        nb_first_visits += send_to_celery(
            backend,
            visit_type_to_queue=visit_type_to_queue,
            policy="first_visits_after_listing",
            lister_name=lister.name,
            lister_instance_name=lister.instance_name,
        )

        def all_first_visits_scheduled():
            nb_listed_origins = 0
            nb_scheduled_origins = 0
            for listed_origins in grouper(
                stream_results(
                    partial(backend.get_listed_origins, lister_id=lister.id)
                ),
                n=1000,
            ):
                listed_origins_list = list(listed_origins)
                nb_listed_origins += len(listed_origins_list)
                for origin_visit_stats in backend.origin_visit_stats_get(
                    ids=(
                        (origin.url, origin.visit_type)
                        for origin in listed_origins_list
                    )
                ):
                    nb_scheduled_origins += 1
                    if (
                        origin_visit_stats.last_scheduled
                        < lister.last_listing_finished_at
                    ):
                        return False
            return nb_scheduled_origins == nb_listed_origins

        if all_first_visits_scheduled():
            # mark that all first visits were scheduled to no longer consider that
            # lister in future execution of that command
            logger.info(
                "All first visits of origins registered by lister with name '%s' "
                "and instance '%s' were scheduled.'",
                lister.name,
                lister.instance_name,
            )
            lister.first_visits_scheduled_at = utcnow()
            backend.update_lister(lister)

    return nb_first_visits



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from datetime import timedelta
import logging
from typing import Tuple

from swh.scheduler.model import ListedOrigin, OriginVisitStats
from swh.scheduler.tests.test_cli import invoke as basic_invoke
from swh.scheduler.utils import utcnow

MODULE_NAMES = [
    "swh.scheduler.cli.admin",
    "swh.scheduler.celery_backend.first_visits",
    "swh.scheduler.celery_backend.utils",
]


def invoke(
    scheduler, args: Tuple[str, ...] = (), catch_exceptions: bool = False, **kwargs
):
    """Invoke swh cli and returns the result call."""
    return basic_invoke(
        scheduler, args=list(args), catch_exceptions=catch_exceptions, **kwargs
    )


def set_log_level(caplog, module_names):
    """Configure the caplog fixtures"""
    for module_name in module_names:
        caplog.set_level(logging.INFO, logger=module_name)


def test_runner_high_priority_first_visits(
    swh_scheduler,
    swh_scheduler_celery_app,
    visit_types,
    task_types,
    mocker,
    caplog,
):
    set_log_level(caplog, MODULE_NAMES)

    nb_origins_per_visit_type = 10
    # number of celery queue slots per visit type is lower than the
    # number of listed origins for that visit type
    nb_available_queue_slots = nb_origins_per_visit_type // 2

    mocker.patch.object(swh_scheduler_celery_app, "send_task")
    # mock number of available slots in queues to a fixed value
    mocker.patch(
        "swh.scheduler.celery_backend.config.get_available_slots"
    ).return_value = nb_available_queue_slots

    # create a lister with high priority first visits
    lister_name = "save-bulk"
    lister_instance_name = "foo"
    lister = swh_scheduler.get_or_create_lister(
        name=lister_name,
        instance_name=lister_instance_name,
        first_visits_queue_prefix="save_bulk",
    )
    # register origins for the lister
    for visit_type in visit_types:
        listed_origins = [
            ListedOrigin(
                lister_id=lister.id,
                url=f"https://{visit_type}.example.org/project{i}",
                visit_type=visit_type,
            )
            for i in range(nb_origins_per_visit_type)
        ]
        swh_scheduler.record_listed_origins(listed_origins)
        # mark some origins visits already scheduled in the past to
        # check they are scheduled again by the tested command
        swh_scheduler.origin_visit_stats_upsert(
            [
                OriginVisitStats(
                    url=listed_origin.url,
                    visit_type=listed_origin.visit_type,
                    last_scheduled=utcnow() - timedelta(days=180),
                )
                for i, listed_origin in enumerate(listed_origins)
                if i % 2 == 1
            ]
        )

    # mark listing as finished
    lister = lister.evolve(last_listing_finished_at=utcnow())
    lister = swh_scheduler.update_lister(lister)

    # start scheduling first visits
    result = invoke(swh_scheduler, args=("start-runner-first-visits",))
    assert result.exit_code == 0

    records = [record.message for record in caplog.records]

    # check expected number of visits were scheduled
    for visit_type in visit_types:
        assert (
            f"{nb_available_queue_slots} visits of type {visit_type} to send to celery"
            in records
        )

    # check there is still origin first visits to schedule
    assert (
        f"All first visits of origins registered by lister with name '{lister_name}' "
        f"and instance '{lister_instance_name}' were scheduled.'"
    ) not in records

    lister = swh_scheduler.get_lister(
        name=lister_name, instance_name=lister_instance_name
    )
    assert lister.first_visits_scheduled_at is None

    # Reset the logs for the next cli call
    caplog.clear()

    # continue scheduling first visits
    result = invoke(swh_scheduler, args=("start-runner-first-visits",))

    assert result.exit_code == 0
    records = [record.message for record in caplog.records]

    # check expected number of visits were scheduled
    for visit_type in visit_types:
        assert (
            f"{nb_available_queue_slots} visits of type {visit_type} to send to celery"
            in records
        )
    # check all listed origins first visits were scheduled
    assert (
        f"All first visits of origins registered by lister with name '{lister_name}' "
        f"and instance '{lister_instance_name}' were scheduled.'"
    ) in records

    lister = swh_scheduler.get_lister(
        name=lister_name, instance_name=lister_instance_name
    )
    assert lister.first_visits_scheduled_at is not None



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from datetime import timedelta
import logging
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from swh.scheduler.interface import SchedulerInterface
    from swh.scheduler.model import TaskType


logger = logging.getLogger(__name__)


def get_loader_task_type(
    scheduler: SchedulerInterface, visit_type: str
) -> Optional[TaskType]:
    "Given a visit type, return its associated task type."
    return scheduler.get_task_type(f"load-{visit_type}")


def send_to_celery(
    scheduler: SchedulerInterface,
    visit_type_to_queue: Dict[str, str],
    enabled: bool = True,
    lister_name: Optional[str] = None,
    lister_instance_name: Optional[str] = None,
    policy: str = "oldest_scheduled_first",
    tablesample: Optional[float] = None,
    absolute_cooldown: Optional[timedelta] = None,
    scheduled_cooldown: Optional[timedelta] = None,
    failed_cooldown: Optional[timedelta] = None,
    not_found_cooldown: Optional[timedelta] = None,
):
    """Utility function to read tasks from the scheduler and send those directly to
    celery.

    Args:
        visit_type_to_queue: Optional mapping of visit/loader type (e.g git, svn, ...)
          to queue to send task to.
        enabled: Determine whether we want to list enabled or disabled origins. As
          default, we want reasonably enabled origins. For some edge case, we might
          want the others.
        lister_name: Determine the list of origins listed from the lister with name
        lister_instance_name: Determine the list of origins listed from the lister
          with instance name
        policy: the scheduling policy used to select which visits to schedule
        tablesample: the percentage of the table on which we run the query
          (None: no sampling)
        absolute_cooldown: the minimal interval between two visits of the same origin
        scheduled_cooldown: the minimal interval before which we can schedule
          the same origin again if it's not been visited
        failed_cooldown: the minimal interval before which we can reschedule a
          failed origin
        not_found_cooldown: the minimal interval before which we can reschedule a
          not_found origin

    Returns:
        The number of tasks sent to celery

    """

    from kombu.utils.uuid import uuid

    from swh.scheduler.celery_backend.config import app, get_available_slots

    from ..utils import create_origin_tasks

    nb_scheduled_tasks = 0
    for visit_type_name, queue_name in visit_type_to_queue.items():
        task_type = get_loader_task_type(scheduler, visit_type_name)
        assert task_type is not None
        task_name = task_type.backend_name
        num_tasks = get_available_slots(app, queue_name, task_type.max_queue_length)

        logger.info("%s slots available in celery queue %s", num_tasks, queue_name)

        origins = scheduler.grab_next_visits(
            visit_type_name,
            num_tasks,
            policy=policy,
            tablesample=tablesample,
            enabled=enabled,
            lister_name=lister_name,
            lister_instance_name=lister_instance_name,
            absolute_cooldown=absolute_cooldown,
            scheduled_cooldown=scheduled_cooldown,
            failed_cooldown=failed_cooldown,
            not_found_cooldown=not_found_cooldown,
        )

        logger.info(
            "%s visits of type %s to send to celery", len(origins), visit_type_name
        )
        for task in create_origin_tasks(origins, scheduler):
            app.send_task(
                task_name,
                task_id=uuid(),
                args=task.arguments.args,
                kwargs=task.arguments.kwargs,
                queue=queue_name,
            )
            nb_scheduled_tasks += 1

    return nb_scheduled_tasks



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This module implements a model of the frequency of updates of an origin
and how long it takes to load it.

For each origin, a commit frequency is chosen deterministically based on the
hash of its URL and assume all origins were created on an arbitrary epoch.
From this we compute a number of commits, that is the product of these two.

And the run time of a load task is approximated as proportional to the number
of commits since the previous visit of the origin (possibly 0)."""

from datetime import datetime, timedelta, timezone
import hashlib
import logging
from typing import Dict, Generator, Iterator, List, Optional, Tuple
import uuid

import attr
from simpy import Event

from swh.model.model import OriginVisitStatus
from swh.scheduler.model import ListedOrigin

from .common import Environment, Queue, Task, TaskEvent

logger = logging.getLogger(__name__)


_nb_generated_origins = 0
_visit_times: Dict[Tuple[str, str], datetime] = {}
"""Cache of the time of the last visit of (visit_type, origin_url),
to spare an SQL query (high latency)."""


def generate_listed_origin(
    lister_id: uuid.UUID, now: Optional[datetime] = None
) -> ListedOrigin:
    """Returns a globally unique new origin. Seed the `last_update` value
    according to the OriginModel and the passed timestamp.

    Arguments:
      lister: instance of the lister that generated this origin
      now: time of listing, to emulate last_update (defaults to :func:`datetime.now`)
    """
    global _nb_generated_origins
    _nb_generated_origins += 1
    assert _nb_generated_origins < 10**6, "Too many origins!"

    if now is None:
        now = datetime.now(tz=timezone.utc)

    url = f"https://example.com/{_nb_generated_origins:06d}.git"
    visit_type = "test-git"
    origin = OriginModel(visit_type, url)

    return ListedOrigin(
        lister_id=lister_id,
        url=url,
        visit_type=visit_type,
        last_update=origin.get_last_update(now),
    )


class OriginModel:
    MIN_RUN_TIME = 0.5
    """Minimal run time for a visit (retrieved from production data)"""

    MAX_RUN_TIME = 7200
    """Max run time for a visit"""

    PER_COMMIT_RUN_TIME = 0.1
    """Run time per commit"""

    EPOCH = datetime(2015, 9, 1, 0, 0, 0, tzinfo=timezone.utc)
    """The origin of all origins (at least according to Software Heritage)"""

    def __init__(self, type: str, origin: str):
        self.type = type
        self.origin = origin

    def seconds_between_commits(self):
        """Returns a random 'average time between two commits' of this origin,
        used to estimate the run time of a load task, and how much the loading
        architecture is lagging behind origin updates."""
        n_bytes = 2
        num_buckets = 2 ** (8 * n_bytes)

        # Deterministic seed to generate "random" characteristics of this origin
        bucket = int.from_bytes(
            hashlib.md5(self.origin.encode()).digest()[0:n_bytes], "little"
        )

        # minimum: 1 second (bucket == 0)
        # max: 10 years (bucket == num_buckets - 1)
        ten_y = 10 * 365 * 24 * 3600

        return ten_y ** (bucket / num_buckets)
        # return 1 + (ten_y - 1) * (bucket / (num_buckets - 1))

    def get_last_update(self, now: datetime) -> datetime:
        """Get the last_update value for this origin.

        We assume that the origin had its first commit at `EPOCH`, and that one
        commit happened every `self.seconds_between_commits()`. This returns
        the last commit date before or equal to `now`.
        """
        _, time_since_last_commit = divmod(
            (now - self.EPOCH).total_seconds(), self.seconds_between_commits()
        )

        return now - timedelta(seconds=time_since_last_commit)

    def get_current_snapshot_id(self, now: datetime) -> bytes:
        """Get the current snapshot for this origin.

        To generate a snapshot id, we calculate the number of commits since the
        EPOCH, and hash it alongside the origin type and url.
        """
        commits_since_epoch, _ = divmod(
            (now - self.EPOCH).total_seconds(), self.seconds_between_commits()
        )
        return hashlib.sha1(
            f"{self.type} {self.origin} {commits_since_epoch}".encode()
        ).digest()

    def load_task_characteristics(
        self, now: datetime
    ) -> Tuple[float, str, Optional[bytes]]:
        """Returns the (run_time, end_status, snapshot id) of the next
        origin visit."""
        current_snapshot = self.get_current_snapshot_id(now)

        key = (self.type, self.origin)
        last_visit = _visit_times.get(key, now - timedelta(days=365))
        time_since_last_successful_run = now - last_visit

        _visit_times[key] = now

        seconds_between_commits = self.seconds_between_commits()
        seconds_since_last_successful = time_since_last_successful_run.total_seconds()
        n_commits = int(seconds_since_last_successful / seconds_between_commits)
        logger.debug(
            "%s characteristics %s origin=%s: Interval: %s, n_commits: %s",
            now,
            self.type,
            self.origin,
            timedelta(seconds=seconds_between_commits),
            n_commits,
        )

        run_time = self.MIN_RUN_TIME + self.PER_COMMIT_RUN_TIME * n_commits
        if run_time > self.MAX_RUN_TIME:
            # Long visits usually fail
            return (self.MAX_RUN_TIME, "partial", None)
        else:
            return (run_time, "full", current_snapshot)


def lister_process(
    env: Environment, lister_id: uuid.UUID
) -> Generator[Event, Event, None]:
    """Every hour, generate new origins and update the `last_update` field for
    the ones this process generated in the past"""
    NUM_NEW_ORIGINS = 100

    origins: List[ListedOrigin] = []

    while True:
        updated_origins = []
        for origin in origins:
            model = OriginModel(origin.visit_type, origin.url)
            updated_origins.append(
                attr.evolve(origin, last_update=model.get_last_update(env.time))
            )

        origins = updated_origins
        origins.extend(
            generate_listed_origin(lister_id, now=env.time)
            for _ in range(NUM_NEW_ORIGINS)
        )

        env.scheduler.record_listed_origins(origins)

        yield env.timeout(3600)


def load_task_process(
    env: Environment, task: Task, status_queue: Queue
) -> Iterator[Event]:
    """A loading task. This pushes OriginVisitStatus objects to the
    status_queue to simulate the visible outcomes of the task.

    Uses the `load_task_duration` function to determine its run time.
    """
    status = OriginVisitStatus(
        origin=task.origin,
        visit=42,
        type=task.visit_type,
        status="created",
        date=env.time,
        snapshot=None,
    )
    logger.debug("%s task %s origin=%s: Start", env.time, task.visit_type, task.origin)
    yield status_queue.put(TaskEvent(task=task, status=status))

    origin_model = OriginModel(task.visit_type, task.origin)
    (run_time, end_status, snapshot) = origin_model.load_task_characteristics(env.time)
    yield env.timeout(run_time)

    logger.debug("%s task %s origin=%s: End", env.time, task.visit_type, task.origin)

    yield status_queue.put(
        TaskEvent(
            task=task,
            status=attr.evolve(
                status, status=end_status, date=env.time, snapshot=snapshot
            ),
        )
    )

    env.report.record_visit(
        (task.visit_type, task.origin), run_time, end_status, snapshot
    )



# Copyright (C) 2018-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import group, shared_task

TASK_PING = "swh.scheduler.tests.tasks.ping"
TASK_MULTIPING = "swh.scheduler.tests.tasks.multiping"
TASK_ERROR = "swh.scheduler.tests.tasks.error"
TASK_ADD = "swh.scheduler.tests.tasks.add"
TASK_ECHO = "swh.scheduler.tests.tasks.echo"


@shared_task(name=TASK_PING, bind=True)
def ping(self, **kw):
    # check this is a SWHTask
    assert hasattr(self, "log")
    assert not hasattr(self, "run_task")
    assert "SWHTask" in [x.__name__ for x in self.__class__.__mro__]
    self.log.debug(self.name)
    if kw:
        return "OK (kw=%s)" % kw
    return "OK"


@shared_task(name=TASK_MULTIPING, bind=True)
def multiping(self, n=10):
    promise = group(ping.s(i=i) for i in range(n))()
    self.log.debug("%s OK (spawned %s subtasks)" % (self.name, n))
    promise.save()
    return promise.id


@shared_task(name=TASK_ERROR)
def not_implemented(*args, **kwargs):
    raise NotImplementedError("Nope")


@shared_task(name=TASK_ADD)
def add(x, y):
    return x + y


@shared_task(name=TASK_ECHO)
def echo(**kw):
    "Does nothing, just return the given kwargs dict"
    return kw



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeVar, Union
from uuid import UUID

import attr
import attr.converters
from attrs_strict import type_validator


def check_timestamptz(value) -> None:
    """Checks the date has a timezone."""
    if value is not None and value.tzinfo is None:
        raise ValueError("date must be a timezone-aware datetime.")


SchedulerModelType = TypeVar("SchedulerModelType", bound="BaseSchedulerModel")


@attr.s
class BaseSchedulerModel:
    """Base class for database-backed objects.

    These database-backed objects are defined through attrs-based attributes
    that match the columns of the database 1:1. This is a (very) lightweight
    ORM.

    These attrs-based attributes have metadata specific to the functionality
    expected from these fields in the database:

     - `primary_key`: the column is a primary key; it should be filtered out
       when doing an `update` of the object
     - `auto_primary_key`: the column is a primary key, which is automatically handled
       by the database. It will not be inserted to. This must be matched with a
       database-side default value.
     - `auto_now_add`: the column is a timestamp that is set to the current time when
       the object is inserted, and never updated afterwards. This must be matched with
       a database-side default value.
     - `auto_now`: the column is a timestamp that is set to the current time when
       the object is inserted or updated.

    """

    _pk_cols: Optional[Tuple[str, ...]] = None
    _select_cols: Optional[Tuple[str, ...]] = None
    _insert_cols_and_metavars: Optional[Tuple[Tuple[str, ...], Tuple[str, ...]]] = None

    @classmethod
    def primary_key_columns(cls) -> Tuple[str, ...]:
        """Get the primary key columns for this object type"""
        if cls._pk_cols is None:
            columns: List[str] = []
            for field in attr.fields(cls):
                if any(
                    field.metadata.get(flag)
                    for flag in ("auto_primary_key", "primary_key")
                ):
                    columns.append(field.name)
            cls._pk_cols = tuple(sorted(columns))

        return cls._pk_cols

    @classmethod
    def select_columns(cls) -> Tuple[str, ...]:
        """Get all the database columns needed for a `select` on this object type"""
        if cls._select_cols is None:
            columns: List[str] = []
            for field in attr.fields(cls):
                columns.append(field.name)
            cls._select_cols = tuple(sorted(columns))

        return cls._select_cols

    @classmethod
    def insert_columns_and_metavars(cls) -> Tuple[Tuple[str, ...], Tuple[str, ...]]:
        """Get the database columns and metavars needed for an `insert` or `update` on
           this object type.

        This implements support for the `auto_*` field metadata attributes.
        """
        if cls._insert_cols_and_metavars is None:
            zipped_cols_and_metavars: List[Tuple[str, str]] = []

            for field in attr.fields(cls):
                if any(
                    field.metadata.get(flag)
                    for flag in ("auto_now_add", "auto_primary_key")
                ):
                    continue
                elif field.metadata.get("auto_now"):
                    zipped_cols_and_metavars.append((field.name, "now()"))
                else:
                    zipped_cols_and_metavars.append((field.name, f"%({field.name})s"))

            zipped_cols_and_metavars.sort()

            cols, metavars = zip(*zipped_cols_and_metavars)
            cls._insert_cols_and_metavars = cols, metavars

        return cls._insert_cols_and_metavars

    def evolve(self: SchedulerModelType, **kwargs) -> SchedulerModelType:
        """Alias to call :func:`attr.evolve` on this object, returning a new object."""
        return attr.evolve(self, **kwargs)

    def to_dict(self: SchedulerModelType, **kwargs) -> Dict[str, Any]:
        """Alias to call :func:`attr.asdict` on this object."""
        return attr.asdict(self, **kwargs)

    def to_tuple(self: SchedulerModelType, **kwargs) -> Tuple[Any]:
        """Alias to call :func:`attr.astuple` on this object."""
        return attr.astuple(self, **kwargs)


@attr.s
class Lister(BaseSchedulerModel):
    name = attr.ib(type=str, validator=[type_validator()])
    instance_name = attr.ib(type=str, validator=[type_validator()])

    # Populated by database
    id = attr.ib(
        type=Optional[UUID],
        validator=type_validator(),
        default=None,
        metadata={"auto_primary_key": True},
    )

    current_state = attr.ib(
        type=Dict[str, Any], validator=[type_validator()], factory=dict
    )
    created = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
        metadata={"auto_now_add": True},
    )
    updated = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
        metadata={"auto_now": True},
    )
    last_listing_finished_at = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
    )
    first_visits_queue_prefix = attr.ib(
        type=Optional[str], validator=[type_validator()], default=None
    )
    first_visits_scheduled_at = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
    )


@attr.s
class ListedOrigin(BaseSchedulerModel):
    """Basic information about a listed origin, output by a lister"""

    lister_id = attr.ib(
        type=UUID, validator=[type_validator()], metadata={"primary_key": True}
    )
    url = attr.ib(
        type=str, validator=[type_validator()], metadata={"primary_key": True}
    )
    visit_type = attr.ib(
        type=str, validator=[type_validator()], metadata={"primary_key": True}
    )
    extra_loader_arguments = attr.ib(
        type=Dict[str, Any], validator=[type_validator()], factory=dict
    )

    last_update = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
    )

    is_fork = attr.ib(type=Optional[bool], validator=[type_validator()], default=None)
    forked_from_url = attr.ib(
        type=Optional[str], validator=[type_validator()], default=None
    )

    enabled = attr.ib(type=bool, validator=[type_validator()], default=True)

    first_seen = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
        metadata={"auto_now_add": True},
    )
    last_seen = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
        metadata={"auto_now": True},
    )


class LastVisitStatus(Enum):
    successful = "successful"
    failed = "failed"
    not_found = "not_found"


def convert_last_visit_status(
    s: Union[None, str, LastVisitStatus]
) -> Optional[LastVisitStatus]:
    if not isinstance(s, str):
        return s
    return LastVisitStatus(s)


@attr.s(frozen=True, slots=True)
class OriginVisitStats(BaseSchedulerModel):
    """Represents an aggregated origin visits view."""

    url = attr.ib(
        type=str, validator=[type_validator()], metadata={"primary_key": True}
    )
    visit_type = attr.ib(
        type=str, validator=[type_validator()], metadata={"primary_key": True}
    )
    last_successful = attr.ib(
        type=Optional[datetime.datetime], validator=type_validator(), default=None
    )
    last_visit = attr.ib(
        type=Optional[datetime.datetime], validator=type_validator(), default=None
    )
    last_visit_status = attr.ib(
        type=Optional[LastVisitStatus],
        validator=type_validator(),
        default=None,
        converter=convert_last_visit_status,
    )
    last_scheduled = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
    )
    last_snapshot = attr.ib(
        type=Optional[bytes], validator=type_validator(), default=None
    )
    next_visit_queue_position = attr.ib(
        type=Optional[int], validator=type_validator(), default=None
    )
    next_position_offset = attr.ib(type=int, validator=type_validator(), default=4)

    successive_visits = attr.ib(type=int, validator=type_validator(), default=1)

    @last_successful.validator
    def check_last_successful(self, attribute, value):
        check_timestamptz(value)

    @last_visit.validator
    def check_last_visit(self, attribute, value):
        check_timestamptz(value)


@attr.s(frozen=True, slots=True)
class SchedulerMetrics(BaseSchedulerModel):
    """Metrics for the scheduler, aggregated by (lister_id, visit_type)"""

    lister_id = attr.ib(
        type=UUID, validator=[type_validator()], metadata={"primary_key": True}
    )
    visit_type = attr.ib(
        type=str, validator=[type_validator()], metadata={"primary_key": True}
    )

    last_update = attr.ib(
        type=Optional[datetime.datetime],
        validator=[type_validator()],
        default=None,
    )

    origins_known = attr.ib(type=int, validator=[type_validator()], default=0)
    """Number of known (enabled or disabled) origins"""

    origins_enabled = attr.ib(type=int, validator=[type_validator()], default=0)
    """Number of origins that were present in the latest listings"""

    origins_never_visited = attr.ib(type=int, validator=[type_validator()], default=0)
    """Number of enabled origins that have never been visited
    (according to the visit cache)"""

    origins_with_pending_changes = attr.ib(
        type=int, validator=[type_validator()], default=0
    )
    """Number of enabled origins with known activity (recorded by a lister)
    since our last visit"""


@attr.s(frozen=True, slots=True)
class TaskType(BaseSchedulerModel):
    """Type of schedulable tasks"""

    type = attr.ib(
        type=str, validator=[type_validator()], metadata={"primary_key": True}
    )
    """Short identifier for the task type"""
    description = attr.ib(type=str, validator=[type_validator()])
    """Human-readable task description"""
    backend_name = attr.ib(type=str, validator=[type_validator()])
    """Name of the task in the job-running backend"""
    default_interval = attr.ib(
        type=Optional[datetime.timedelta], validator=[type_validator()], default=None
    )
    """Default interval for newly scheduled tasks"""
    min_interval = attr.ib(
        type=Optional[datetime.timedelta], validator=[type_validator()], default=None
    )
    """Minimum interval between two runs of a task"""
    max_interval = attr.ib(
        type=Optional[datetime.timedelta], validator=[type_validator()], default=None
    )
    """Maximum interval between two runs of a task"""
    backoff_factor = attr.ib(
        type=Optional[float], validator=[type_validator()], default=None
    )
    """Adjustment factor for the backoff between two task runs"""
    max_queue_length = attr.ib(
        type=Optional[int], validator=[type_validator()], default=1000
    )
    """Maximum length of the queue for this type of tasks, default to 1000 if not provided"""
    num_retries = attr.ib(
        type=Optional[int], validator=[type_validator()], default=None
    )
    """Default number of retries on transient failures"""
    retry_delay = attr.ib(
        type=Optional[datetime.timedelta], validator=[type_validator()], default=None
    )
    """Retry delay for the task"""


@attr.s(frozen=True, slots=True)
class TaskArguments(BaseSchedulerModel):
    args = attr.ib(type=List[Any], validator=[type_validator()], default=[])
    kwargs = attr.ib(type=Dict[str, Any], validator=[type_validator()], default={})


TaskStatus = Literal[
    "next_run_not_scheduled", "next_run_scheduled", "completed", "disabled"
]

TaskPolicy = Literal["recurring", "oneshot"]

TaskPriority = Literal["high", "normal", "low"]


@attr.s(frozen=True, slots=True)
class Task(BaseSchedulerModel):
    """Represents a schedulable task"""

    type = attr.ib(type=str, validator=[type_validator()])
    """Task type"""
    arguments = attr.ib(type=TaskArguments, validator=[type_validator()])
    """Task arguments passed to the underlying job scheduler"""
    next_run = attr.ib(type=datetime.datetime, validator=[type_validator()])
    """The interval between two runs of this task taking into account the backoff factor"""
    status = attr.ib(
        type=TaskStatus, validator=[type_validator()], default="next_run_not_scheduled"
    )
    """Status of the task"""
    policy = attr.ib(type=TaskPolicy, validator=[type_validator()], default="recurring")
    """Whether the task is one-shot or recurring"""
    retries_left = attr.ib(
        type=Optional[int], validator=[type_validator()], default=None
    )
    """The number of "short delay" retries of the task in case of transient failure"""
    id = attr.ib(
        type=Optional[int],
        validator=[type_validator()],
        metadata={"primary_key": True},
        default=None,
    )
    """Task Identifier (populated by database)"""
    current_interval = attr.ib(
        type=Optional[datetime.timedelta], validator=[type_validator()], default=None
    )
    """The next run of this task should be run on or after that time"""
    priority = attr.ib(
        type=Optional[TaskPriority], validator=[type_validator()], default=None
    )
    """Priority of the task, either low, normal or high"""


TaskRunStatus = Literal[
    "scheduled", "started", "eventful", "uneventful", "failed", "permfailed", "lost"
]


@attr.s(frozen=True, slots=True)
class TaskRun(BaseSchedulerModel):
    """Represents the execution of a task sent to the job-running backend"""

    task = attr.ib(type=Optional[int], validator=[type_validator()])
    """Task identifier"""
    id = attr.ib(
        type=Optional[int],
        validator=[type_validator()],
        metadata={"primary_key": True},
        default=None,
    )
    """Task run identifier (populated by database)"""
    backend_id = attr.ib(type=Optional[str], validator=[type_validator()], default=None)
    """id of the task run in the job-running backend"""
    scheduled = attr.ib(
        type=Optional[datetime.datetime], validator=[type_validator()], default=None
    )
    """Scheduled run time for task"""
    started = attr.ib(
        type=Optional[datetime.datetime], validator=[type_validator()], default=None
    )
    """Task starting time"""
    ended = attr.ib(
        type=Optional[datetime.datetime], validator=[type_validator()], default=None
    )
    """Task ending time"""
    metadata = attr.ib(
        type=Optional[Dict[str, Any]], validator=[type_validator()], default=None
    )
    """Useful metadata for the given task run. For instance, the worker
    that took on the job, or the logs for the run"""
    status = attr.ib(
        type=TaskRunStatus, validator=[type_validator()], default="scheduled"
    )
    """Status of the task run"""



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
from datetime import datetime
import random
from typing import Dict, List, Optional, Tuple

import attr

from swh.scheduler.interface import SchedulerInterface
from swh.scheduler.model import LastVisitStatus, OriginVisitStats

msg_type = "origin_visit_status"

DISABLE_ORIGIN_THRESHOLD = 3
"""Threshold to disable failing origins"""

MAX_NEXT_POSITION_OFFSET = 10
"""Max next position offset to avoid date computation overflow"""


def max_date(*dates: Optional[datetime]) -> datetime:
    """Return the max date of given (possibly None) dates

    At least one date must be not None.
    """
    filtered_dates = [d for d in dates if d is not None]
    if not filtered_dates:
        raise ValueError("At least one date should be a valid datetime")

    return max(filtered_dates)


def from_position_offset_to_days(position_offset: int) -> int:
    """Compute position offset to interval in days. Note that this does not bound the
       position_offset input so client code should limit the date computation to avoid
       overflow errors.

        - index in [0:1]: interval 1 day
        - index in [2:4]: interval 2 days
        - index in [5:+inf]: interval `4^(index-4)` days

    Args:
        position_offset: The actual position offset for a given visit stats

    Returns:
        The offset as an interval in number of days.

    """
    assert position_offset >= 0
    if position_offset < 2:
        result = 1
    elif position_offset < 5:
        result = 2
    else:
        result = 4 ** (position_offset - 4)
    return result


def next_visit_queue_position(
    queue_position_per_visit_type: Dict[str, int], visit_stats: Dict
) -> int:
    """Compute the next visit queue position for the given visit_stats.

    This takes the visit_stats next_position_offset value and compute a corresponding
    interval in "days" (with a random fudge factor of -/+ 10% range to avoid scheduling
    burst for hosters). Then computes out of this visit interval and the current visit
    stats's position in the queue a new position.

    As an implementation detail, if the visit stats does not have a queue position yet,
    this fallbacks to use the current global position (for the same visit type as the
    visit stats) to compute the new position in the queue. If there is no global state
    yet for the visit type, this starts up using 0 as default value.

    Args:
        queue_position_per_visit_type: The global state of the queue per visit type
        visit_stats: The actual visit information to compute the next position for

    Returns:
        The actual next visit queue position for that visit stats

    """
    days = from_position_offset_to_days(visit_stats["next_position_offset"])
    random_fudge_factor = random.uniform(-0.1, 0.1)
    visit_interval = int(days * 24 * 3600 * (1 + random_fudge_factor))
    # Use the current queue position per visit type as starting position if none is
    # already set
    default_queue_position = queue_position_per_visit_type.get(
        visit_stats["visit_type"], 0
    )

    current_position = (
        visit_stats["next_visit_queue_position"]
        if visit_stats.get("next_visit_queue_position")
        else default_queue_position
    )
    return current_position + visit_interval


def get_last_status(
    incoming_visit_status: Dict, known_visit_stats: Dict
) -> Tuple[LastVisitStatus, Optional[bool]]:
    """Determine the `last_visit_status` and eventfulness of an origin according to
    the received visit_status object, and the state of the origin_visit_stats in db.

    Note that at the time this function is called, out of order messages were already
    discarded. Thus why the implementation is rather simple.

    Args:
        incoming_visit_status: Incoming visit status read ouf of the journal
        known_visit_stats: Visit stats already registered in the backend

    Returns:
        A tuple of (LastVisitStatus, Optional[bool]). LastVisitStatus represents the
        successfulness of the visit. Optional[bool] represents whether the snapshot is
        fresher than before (True/False) or None if there is no snapshot at all.

    """

    status = incoming_visit_status["status"]
    if status in ("not_found", "failed"):
        return LastVisitStatus(status), None

    assert status in ("full", "partial")

    if incoming_visit_status["snapshot"] is None:
        return LastVisitStatus.failed, None

    if incoming_visit_status["snapshot"] != known_visit_stats.get("last_snapshot"):
        return LastVisitStatus.successful, True

    return LastVisitStatus.successful, False


def process_journal_objects(
    messages: Dict[str, List[Dict]], *, scheduler: SchedulerInterface
) -> None:
    f"""Read messages from origin_visit_status journal topic to update "origin_visit_stats"
    information on (origin, visit_type). The goal is to compute visit stats information
    per origin and visit_type: `last_successful`, `last_visit`, `last_visit_status`, ...

    Details:

        - This journal consumes origin visit status information for final visit
          status (`"full"`, `"partial"`, `"failed"`, `"not_found"`). It drops
          the information of non final visit statuses (`"ongoing"`,
          `"created"`).

        - This journal client only considers messages that arrive in
          chronological order. Messages that arrive out of order (i.e. with a
          date field smaller than the latest recorded visit of the origin) are
          ignored. This is a tradeoff between correctness and simplicity of
          implementation [1]_.

        - The snapshot is used to determine the eventful or uneventful nature of
          the origin visit.

        - When no snapshot is provided, the visit is considered as failed.

        - Finally, the `next_visit_queue_position` (position in the global per-origin
          type queue at which some new objects are expected to be added for the origin),
          and `next_position_offset` (duration that we expect to wait between visits of
          this origin) are updated.

        - When visits fails at least {DISABLE_ORIGIN_THRESHOLD} times in a row, the
          origins are disabled in the scheduler table. It's up to the lister to activate
          those back when they are listed again.

    This is a worker function to be used with `JournalClient.process(worker_fn)`, after
    currification of `scheduler` and `task_names`.

    .. [1] Ignoring out of order messages makes the initialization of the
      origin_visit_status table (from a full journal) less deterministic: only the
      `last_visit`, `last_visit_state` and `last_successful` fields are guaranteed
      to be exact, the `next_position_offset` field is a best effort estimate
      (which should converge once the client has run for a while on in-order
      messages).

    """
    assert set(messages) <= {
        msg_type
    }, f"Got unexpected {', '.join(set(messages) - set([msg_type]))} message types"
    assert msg_type in messages, f"Expected {msg_type} messages"

    interesting_messages = [
        msg
        for msg in messages[msg_type]
        if "type" in msg and msg["status"] not in ("created", "ongoing")
    ]

    if not interesting_messages:
        return

    origin_visit_stats: Dict[Tuple[str, str], Dict] = {
        (visit_stats.url, visit_stats.visit_type): visit_stats.to_dict()
        for visit_stats in scheduler.origin_visit_stats_get(
            list(set((vs["origin"], vs["type"]) for vs in interesting_messages))
        )
    }
    existing_origin_visit_stats = copy.deepcopy(origin_visit_stats)

    # Use the default values from the model object
    empty_object = {
        field.name: field.default if field.default != attr.NOTHING else None
        for field in attr.fields(OriginVisitStats)
    }

    disabled_urls: List[str] = []

    # Retrieve the global queue state
    queue_position_per_visit_type = scheduler.visit_scheduler_queue_position_get()

    for msg_dict in interesting_messages:
        origin = msg_dict["origin"]
        visit_type = msg_dict["type"]
        pk = origin, visit_type
        if pk not in origin_visit_stats:
            origin_visit_stats[pk] = {
                **empty_object,
                "url": origin,
                "visit_type": visit_type,
            }

        visit_stats_d = origin_visit_stats[pk]

        if (
            visit_stats_d.get("last_visit")
            and msg_dict["date"] <= visit_stats_d["last_visit"]
        ):
            # message received out of order, ignore
            continue

        # Compare incoming message to known status of the origin, to determine
        # eventfulness
        last_visit_status, eventful = get_last_status(msg_dict, visit_stats_d)

        # Update the position offset according to the visit status,
        # if we had already visited this origin before.

        if visit_stats_d.get("last_visit"):
            # Update the next position offset according to the existing value and the
            # eventfulness of the visit.
            increment = -2 if eventful else 1
            # Limit the next_position_offset for acceptable date computations
            current_offset = min(
                visit_stats_d["next_position_offset"] + increment,
                MAX_NEXT_POSITION_OFFSET,
            )
            visit_stats_d["next_position_offset"] = max(0, current_offset)
            # increment the counter when last_visit_status is the same
            same_visit_status = last_visit_status == visit_stats_d["last_visit_status"]
        else:
            same_visit_status = False

        # Record current visit date as highest known date (we've rejected out of order
        # messages earlier).
        visit_stats_d["last_visit"] = msg_dict["date"]
        visit_stats_d["last_visit_status"] = last_visit_status

        # Record last successful visit date
        if last_visit_status == LastVisitStatus.successful:
            visit_stats_d["last_successful"] = max_date(
                msg_dict["date"], visit_stats_d.get("last_successful")
            )
            visit_stats_d["last_snapshot"] = msg_dict["snapshot"]

        # Update the next visit queue position (which will be used solely for origin
        # without any last_update, cf. the dedicated scheduling policy
        # "origins_without_last_update")
        visit_stats_d["next_visit_queue_position"] = next_visit_queue_position(
            queue_position_per_visit_type, visit_stats_d
        )

        visit_stats_d["successive_visits"] = (
            visit_stats_d["successive_visits"] + 1 if same_visit_status else 1
        )

        # Disable recurring failing/not-found origins
        if (
            visit_stats_d["last_visit_status"]
            in [LastVisitStatus.not_found, LastVisitStatus.failed]
        ) and visit_stats_d["successive_visits"] >= DISABLE_ORIGIN_THRESHOLD:
            disabled_urls.append(visit_stats_d["url"])

    # Only upsert changed values
    to_upsert = []
    for key, ovs in origin_visit_stats.items():
        if (
            key not in existing_origin_visit_stats
            or ovs != existing_origin_visit_stats[key]
        ):
            to_upsert.append(OriginVisitStats(**ovs))

    if to_upsert:
        scheduler.origin_visit_stats_upsert(to_upsert)

    # Disable any origins if any
    if disabled_urls:
        disabled_origins = []
        for url in disabled_urls:
            origins = scheduler.get_listed_origins(url=url).results
            if len(origins) > 0:
                origin = attr.evolve(origins[0], enabled=False)
                disabled_origins.append(origin)

        if disabled_origins:
            scheduler.record_listed_origins(disabled_origins)



# Copyright (C) 2023 the Software Heritage developers
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


class FooLister:
    pass


def register():
    return {
        "lister": FooLister,
        "task_modules": ["%s.tasks" % __name__],
    }



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.scheduler.cli.origin_utils import get_scheduler_task_type


def test_get_scheduler_task_type_found(swh_scheduler, task_types):
    """It should find standard task types"""
    # Check we found the standard task type
    for task_type, task_type_info in task_types.items():
        actual_task_info = get_scheduler_task_type(swh_scheduler, task_type)
        assert actual_task_info == task_type_info


@pytest.mark.parametrize(
    "task_type_suffix", ["bitbucket", "extra-mercurial", "overly-long-task-suffix"]
)
def test_get_scheduler_task_type_found_derivative(
    swh_scheduler, task_types, task_type_suffix
):
    """It should find derivative task types"""
    # Create another task type derivative out of a standard task type
    first_task_type = next(iter(task_types))
    another_task_type = f"{first_task_type}-{task_type_suffix}"

    actual_task_info_2 = get_scheduler_task_type(swh_scheduler, another_task_type)
    # we found it and it matches the origin task type it derives from
    assert actual_task_info_2 == task_types[first_task_type]


def test_get_scheduler_task_type_found_raise(swh_scheduler, task_types):
    """It should raise when no task_type is found"""
    unknown_task_type = "foobar-task-type"
    assert unknown_task_type not in task_types

    with pytest.raises(ValueError, match="not find scheduler"):
        get_scheduler_task_type(swh_scheduler, unknown_task_type)



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This package runs the scheduler in a simulated environment, to evaluate
various metrics. See :ref:`swh-scheduler-simulator`.

This module orchestrates of the simulator by initializing processes and connecting
them together; these processes are defined in modules in the package and
simulate/call specific components."""

from datetime import datetime, timedelta, timezone
import logging
from typing import Dict, Generator, Optional

from simpy import Event

from swh.scheduler.interface import SchedulerInterface
from swh.scheduler.utils import create_origin_tasks, utcnow

from . import origin_scheduler, task_scheduler
from .common import Environment, Queue, SimulationReport, Task
from .origins import generate_listed_origin, lister_process, load_task_process

logger = logging.getLogger(__name__)


def update_metrics_process(
    env: Environment, update_interval: int
) -> Generator[Event, None, None]:
    """Update the scheduler metrics every `update_interval` (simulated) seconds,
    and add them to the SimulationReport
    """
    t0 = env.time
    while True:
        metrics = env.scheduler.update_metrics(timestamp=env.time)
        env.report.record_metrics(env.time, metrics)
        dt = env.time - t0
        logger.info("time:%s visits:%s", dt, env.report.total_visits)
        yield env.timeout(update_interval)


def worker_process(
    env: Environment, name: str, task_queue: Queue, status_queue: Queue
) -> Generator[Event, Task, None]:
    """A worker which consumes tasks from the input task_queue. Tasks
    themselves send OriginVisitStatus objects to the status_queue."""
    logger.debug("%s worker %s: Start", env.time, name)
    while True:
        task = yield task_queue.get()
        logger.debug(
            "%s worker %s: Run task %s origin=%s",
            env.time,
            name,
            task.visit_type,
            task.origin,
        )
        yield env.process(load_task_process(env, task, status_queue=status_queue))


def setup(
    env: Environment,
    scheduler_type: str,
    policy: Optional[str],
    workers_per_type: Dict[str, int],
    task_queue_capacity: int,
    min_batch_size: int,
    metrics_update_interval: int,
):
    task_queues = {
        visit_type: Queue(env, capacity=task_queue_capacity)
        for visit_type in workers_per_type
    }
    status_queue = Queue(env)

    if scheduler_type == "origin_scheduler":
        if policy is None:
            raise ValueError("origin_scheduler needs a scheduling policy")
        env.process(
            origin_scheduler.scheduler_runner_process(
                env, task_queues, policy, min_batch_size=min_batch_size
            )
        )
        env.process(
            origin_scheduler.scheduler_journal_client_process(env, status_queue)
        )
    elif scheduler_type == "task_scheduler":
        if policy is not None:
            raise ValueError("task_scheduler doesn't support a scheduling policy")
        env.process(
            task_scheduler.scheduler_runner_process(
                env, task_queues, min_batch_size=min_batch_size
            )
        )
        env.process(task_scheduler.scheduler_listener_process(env, status_queue))
    else:
        raise ValueError(f"Unknown scheduler type to simulate: {scheduler_type}")

    env.process(update_metrics_process(env, metrics_update_interval))

    for visit_type, num_workers in workers_per_type.items():
        task_queue = task_queues[visit_type]
        for i in range(num_workers):
            worker_name = f"worker-{visit_type}-{i}"
            env.process(worker_process(env, worker_name, task_queue, status_queue))

    lister = env.scheduler.get_or_create_lister(name="example")
    assert lister.id
    env.process(lister_process(env, lister.id))


def fill_test_data(scheduler: SchedulerInterface, num_origins: int = 100000):
    """Fills the database with mock data to test the simulator."""
    stored_lister = scheduler.get_or_create_lister(name="example")
    assert stored_lister.id is not None

    # Generate 'num_origins' new origins
    origins = [generate_listed_origin(stored_lister.id) for _ in range(num_origins)]
    scheduler.record_listed_origins(origins)

    scheduler.create_tasks(
        [
            task.evolve(
                policy="recurring",
                next_run=origin.last_update or utcnow(),
                current_interval=timedelta(days=64),
            )
            for (origin, task) in zip(origins, create_origin_tasks(origins, scheduler))
        ]
    )


def run(
    scheduler: SchedulerInterface,
    scheduler_type: str,
    policy: Optional[str],
    runtime: Optional[int],
):
    NUM_WORKERS = 48
    start_time = datetime.now(tz=timezone.utc)
    env = Environment(start_time=start_time)
    env.scheduler = scheduler
    env.report = SimulationReport()
    setup(
        env,
        scheduler_type=scheduler_type,
        policy=policy,
        workers_per_type={"git": NUM_WORKERS},
        task_queue_capacity=10000,
        min_batch_size=1000,
        metrics_update_interval=3600,
    )
    try:
        env.run(until=runtime)
    except KeyboardInterrupt:
        pass
    finally:
        end_time = env.time
        print("total simulated time:", end_time - start_time)
        metrics = env.scheduler.update_metrics(timestamp=end_time)
        env.report.record_metrics(end_time, metrics)
        return env.report
